# Description: handle "inner" code that has no further nesting. 
# Called from processTexRange().
# Arguments:
#   tex: a block of tex with no further nesting (following shinyTex rules)
#   eNum: the element number of this 'tex' in the doList (used for constructing unique filenames)
#
# Value: a list containing the generated 'uiCode' and 'serverCode' for this block of tex
#
# Side effects: Files named 'eNum'-#.html (e.g., 13-1.html through 13-4.html) are created.
#   These contain pure html code and each have a corresponding includeHTML() Shiny command in 
#   the 'uiCode'.
#
# Details: 
# The following tasks are performed:
#   ?? $math tex$ and $$math tex$$ are converted to mathJax formats
#   ?? LaTex \mathbf, \mathem, \bf, \em and \sc are converted to equivalent HTML
#   * shinyTex commands \shinyTip and \shinyLink (with or without a "tip") are
#     converted to Shiny/HTML equivalents 
#   * verbatim environments are converted to <pre> and </pre> in HTML
#   * "inner" shinyTex commands like shinySlider, shinyRadioButtons, shinyVideo, shinyAudio,
#     and shinyActionButton are converted to Shiny code
#   * "inner" shinyTex environments like shinyPlot are converted to ui+server shiny code
# Note: uses htmlify() and processCommands()  
#
innerCode = function(tex, eNum, before, globalInfoEnv) {
  lineNumOffset = before + get("lineNumOffset", envir=globalInfoEnv)

  ## segment into blocks of text vs. commands vs. special html one-liners vs. inner environments
  innerCommands = c("shinySlider", "shinyRadioButtons", "shinyCheckbox", "shinyTextInput",
                    "shinyTextInputEnter", "shinyTextAreaInput", "shinyActionButton", "shinyQuiz", 
                    "shinyVideo", "shinyAudio", "shinyImage",
                    "shinyNumericInput", "shinyDropdown", "shinyInternalLoc", "shinyInternalLink",
                    "shinyGotoTab")
  innerCommandOrs = paste(innerCommands, collapse="|")
  innerHtml = c("hr", "smallskip", "medskip", "bigskip")
  innerHtmlOrs = paste(innerHtml, collapse="|")
  
  # expand out tabular to shinyTabular
  tex = gsub(paste0('[[:blank:]]*[\\](begin|end)[[:blank:]]*[{]tabular[}]'), '\\\\\\1{shinyTabular}', tex)
  
  # Main work of finding blocks of one or more lines of different kinds to process
  # (These are actually 'shinyPlot', etc.)
  innerEnvirons = c("Plot", "Tabular", "CodeOutput", "Reactive", "Code", "UI", "Server", 
                    "ConditionalText")
  innerEnvironsOrs = paste(innerEnvirons, collapse="|") 
  comRE = regexpr(paste0('[\\](', innerCommandOrs, ')'), tex)
  blockType = rep("f", length(tex)) # for free tex (--> html)
  temp = which(comRE != -1)
  if (length(temp) > 0) {
    for (cmdNum in 1:length(temp)) {
      blockType[temp] = "c" # for command
      # Allow multi-line commands
      for (i in 1:length(temp)) {
        if (length(grep("[}]", tex[temp[i]])) > 0) next
        tempR = paste(tex[temp[i]:length(tex)], collapse="")
        tempB = regexpr("[}]", tempR)
        if (tempB == -1) {
          stop("@", temp[i]+lineNumOffset, " no ending '}' for \\shinyXXX" )
        }
        endLine = min(which(tempB <= cumsum(nchar(tex[temp[i]:length(tex)])))) - 1
        blockType[temp[i]:(temp[i]+endLine)] = "c" # for multi-line command
      }
    } # end for each \shinyActionButton/RadioButtons/Slider/Checkbox
  }
  
  ## Avoid \href catching as \hr
  htmlRE = regexpr(paste0('[\\](', innerHtmlOrs, ')([[:blank:]]|[%])'), paste0(tex, " "))
  blockType[htmlRE != -1] = "h" # for html

  # find and verify inner environment matches
  inners = str_match(tex, paste0('[\\\\](begin|end)[[:blank:]]*[{]shiny(',
                                 innerEnvironsOrs, ')[}][[:blank:]]*',
                                 '([{]([^}]*)[}])?[[:blank:]]*([{]([^}]*)[}])?'))
  innersLines = which(!is.na(inners[,1]))
  inners = inners[!is.na(inners[,1]), , drop=FALSE]
  nri = nrow(inners)
  if (nri > 0) { 
    odd = nri%%2 == 1
    if (odd && inners[nri, 2]=="begin") {
      stop("@", innersLines[1]+lineNumOffset, " no matching 'end' for ", inners[nri, 1])
    }
    if (nri == 1) {
      stop("@", innersLines[1]+lineNumOffset, " no matching 'begin' for ", inners[1, 1])
    }
    if (odd) {
      saved = inners[nri, 1]
      inners = inners[-nri, , drop=FALSE]
      nri = nri - 1
    }
    temp = inners[seq(1, nri, 2), 2]
    temp2 = which(temp != "begin")
    if (length(temp2) > 0) {
      stop("@", innersLines[temp2][1]+lineNumOffset, " no matching 'begin' for ", inners[2*temp2[1]-1, 1])
    }
    temp = inners[seq(2, nri, 2), 2]
    temp2 = which(temp != "end")
    if (length(temp2) > 0) {
      stop("@", innersLines[temp2][1]+lineNumOffset, " no matching end for ", inners[2*temp2[1], 1])
    }
    temp = which(apply(matrix(inners[,3], nrow=2), 2, 
                       function(x) x[1]!=x[2]))
    if (length(temp) > 0) {
      stop("@", innersLines[2*temp[1]-1]+lineNumOffset, "/" , innersLines[2*temp[1]]+lineNumOffset, " ",
           inners[2*temp[1]-1, 1], " improperly paired with ", inners[2*temp[1], 1])
    }
    if (odd) {
      stop("@", innersLines[nri+1]+lineNumOffset, " no matching begin for ", saved)
    }
    for (i in 1:(nri/2)) blockType[innersLines[2*i-1]:innersLines[2*i]] = as.character(i) # environment # i
  } # end marking environment block numbers (vs. "f" or "c")
  
  # Identify and process blocks of code by type/function
  rleBlocks = rle(blockType)
  rleBlockEnd = cumsum(rleBlocks$lengths)
  rleBlockN = length(rleBlockEnd)
  if (rleBlockN==1) {
    rleBlockStart=1 
  } else {
    rleBlockStart=c(1, rleBlockEnd[1:(rleBlockN-1)] + 1)
  }
  rleBlockType = rleBlocks$value
  
  # UI code for commands directly and text as a separate file plus inputHTML UI code
  uiCode = serverCode = NULL
  sub = 0  # subcode for .html text file names
  #
  for (block in 1:rleBlockN) {
    text = tex[rleBlockStart[block] : rleBlockEnd[block]]
    
    # Process a block that is text (may contain equations and/or \textbf{}, etc.)
    if (rleBlockType[block] == "f") {
      finalCheck = regexpr(paste0("(", "[\\]shiny[^[:blank:][:punct:]]+", "|",
                                  "[\\](begin|end)[[:blank:]]*[{]shiny[^[:blank:][:punct:]]+[}]",
                                  ")"),
                           text)
      if (any(finalCheck != -1)) {
        temp = regmatches(text, finalCheck)
        warning("unexpected Shiny command(s)/environment(s): ", paste(temp, collapse="; "), collapse="")
      }
      #
      html = htmlify(text, rleBlockStart[block]+lineNumOffset)
      #
      if (!is.null(html)) {
        sub = sub + 1
        fname = paste0(eNum, "-", sub, ".html")
        write(html, file=fname, ncolumns=1)
        uiCode = c(uiCode, paste0('includeHTML("', fname, '"),'))
      }
      
    # Process a block containing one or more Shiny commands
    } else if (rleBlockType[block]=="c") {
      assure(length(grep("\\\\[[:blank:]]*$",text))==0, 92871)
      # some lines in a block may be multi-line command
      tempN = grep(paste0('[\\](', innerCommandOrs, ')'), text)
      assure(length(tempN) != 0 && tempN[1]==1, 01002)
      tempL = length(text)
      if (length(tempN)!=1 || tempL!=1) {
        if (length(tempN)==1) {
          text = paste(text, collapse="")
        } else {
          mat = cbind(tempN, c(tempN[2:(length(tempN))]-1, tempL))
          text = sapply(1:nrow(mat), function(i) paste0(text[mat[i,1]:mat[i,2]], collapse=""))
        }
      }
      for (line in 1:length(text)) { #### update lineNumOffset to account for multi-line commands
        uiCode = c(uiCode, processCommands(text[line], before + rleBlockStart[block]-1+line, globalInfoEnv))
      }

    ### Process a block with special direct-to-html commands 
    } else if (rleBlockType[block]=="h") { 
      for (line in 1:length(text)) {
        if (length(grep("[\\]hr", text[line])) > 0) {
          text[line] = "hr(),"
        } else if (length(grep("[\\]smallskip", text[line])) > 0) {
          text[line] = "HTML('<div style=\"padding:1ex\"></div>'),"
        } else if (length(grep("[\\]medskip", text[line])) > 0) {
          text[line] = "HTML('<div style=\"padding:2ex\"></div>'),"
        } else if (length(grep("[\\]bigskip", text[line])) > 0) {
          text[line] = "HTML('<div style=\"padding:3ex\"></div>'),"
        } else {
          cat("un-handled shinyHtml: '", text[line], "'\n")
          assure(FALSE, 33121)
        }
      }
      uiCode = c(uiCode, text)
      
    # Process a block that is an inner environment  ## ??? incorporate all options here!!!!!!!!!!!! ????
    # Includes shinyPlot, (shiny)Tabular, shinyReactive, shinyConditionalText, shinyCode, etc.
    } else if (substring(rleBlockType[block],1,1) %in% as.character(0:9)) {
      temp = as.numeric(rleBlockType[block]) * 2 - 1
      command = inners[temp, 3]
      opt1 = inners[temp, 5]
      if (is.na(opt1)) opt1 = ""
      opt2 = inners[temp, 7]
      if (is.na(opt2)) opt2 = ""
      
      if (command == "Plot") {
        ## ????????? !!!!!!!!!! find a better way to generate a plot outputId
        inners[temp, 5] = paste(LETTERS[sample(1:26,6,replace=TRUE)], collapse="")
        if (inners[temp ,5] == "") {
          stop("@", rleBlockStart[block]+lineNumOffset, " need a plot id in braces after \\begin{shinyPlot}")
        }
        uiCode = c(uiCode, paste0('plotOutput(outputId="', inners[temp, 5], '"),'))
        serverCode = c(serverCode, paste0('output$', inners[temp, 5], '=renderPlot({'),
                       paste0("  ", text[2:(length(text)-1)]),
                       '})', '') 
        
      } else if (command == "Tabular") {
        # eliminate \sc and cannot handle \\ or \hline for now, also <em> and <strong>
        data = text[2:(length(text)-1)]
        data = gsub("[{][[:blank:]]*[\\](em|bf|sc)(.*?)[}]", "\\2", data)
        data = gsub("(\\\\|\\hline)", "", data)
        temp = strsplit(data, "&")
        if (length(temp) ==1 ) stop("@", rleBlockStart[block]+lineNumOffset, " 1-row tables not allowed")
        tempL = sapply(temp, length)
        if (!all(tempL == tempL[1])) {
          stop("@", rleBlockStart[block]+lineNumOffset, 
               " table does not have the same number of columns for each row")
        }
        html = c('<table id="shinyTexTable">',
                 paste0('<tr> <th>', paste0(temp[[1]], collapse="</th> <th>"), "</th> </tr>"))
        for (i in 2:length(tempL)) {
          html = c(html, paste0('<tr> <td>', paste0(temp[[i]], collapse="</td> <td>"), "</td> </tr>"))
        }
        html = c(html, '</table>')
        sub = sub + 1
        fname = paste0(eNum, "-", sub, ".html")
        write(html, file=fname, ncolumns=1)
        uiCode = c(uiCode, paste0('includeHTML("', fname, '"),'))

      } else if (command == "UI") {
        if (opt1 !="" || opt2 !="") stop("\\begin{shinyUI} takes no options: ", inners[temp,1])
        temp = text[2:(length(text)-1)]
        while (length(temp)>1 && length(grep("^[[:blank:]]*$", temp[length(temp)]))==1) temp = temp[-length(temp)]
        if (length(temp)==1 && str_trim(temp)=="") {
          warning("@", rleBlockStart[block]+lineNumOffset, " empty shinyUI environment")
        } else {
          temp[length(temp)] = paste0(temp[length(temp)], ",")
          uiCode = c(uiCode, temp)
        }
        
      } else if (command == "Server") {
        if (opt1 !="" || opt2 !="") stop("\\begin{shinyServer} takes no options: ", inners[temp,1])
        temp = text[2:(length(text)-1)]
        while (length(temp)>1 && length(grep("^[[:blank:]]*$", temp[length(temp)]))==1) temp = temp[-length(temp)]
        if (length(temp)==1 && str_trim(temp)=="") {
          warning("@", rleBlockStart[block]+lineNumOffset, " empty shinyServer environment")
        } else {
          temp[length(temp)] = temp[length(temp)]
          serverCode = c(serverCode, temp)
          uiCode = c(uiCode, "p(''),")
        }
        
      } else if (command == "Reactive") {
        if (opt2!="") stop("\\begin{shinyReactive} takes only one option: ", inners[temp,1])
        if (opt1  == "") stop("\\begin{shinyReactive} requires {varName}: ", inners[temp,1])
        if (opt1 != make.names(opt1)) stop("invalid name in \\begin{shinyReactive}{varName}: ", inners[temp,1])
        serverCode = c(serverCode, 
                       paste0(opt1, " = reactive({"),
                       text[2:(length(text)-1)], 
                       "})",
                       '')
        
      } else if (command == "Code") {
        if (opt1 != "" || opt2 !="") stop("\\shinyCode takes no options: ", inners[temp,1])
        serverCode = c(serverCode, 
                       text[2:(length(text)-1)], 
                       '') 
        
      } else if (command == "CodeOutput") {
        textName = paste0("text", round(runif(1,1,99999)))
        uiCode = c(uiCode, paste0('verbatimTextOutput(outputId="', textName, '"),'))
        serverCode = c(serverCode, paste0('output$', textName, '=renderPrint({'),
                       text[2:(length(text)-1)],
                       '})', '')
      } else {
        stop("innerCode() does not yet handle \\shiny", command)
      } # end if command is each of the possible commands
      
    # block code neither "f" nor "c" nor "h" nor a number (environment section)
    } else {
      cat("rleBlockType[block] = ", rleBlockType[block], "\n")
      assure(FALSE, 89756)
    } # end handling each type of code block 
    
  } # end for each block of inner text or shiny commands
  if (is.null(serverCode)) serverCode = NA
  return(list(uiCode=uiCode, serverCode=serverCode))
} # end innerCode()
